import pytest
from mock import patch
from transifex.native.tools.migrations.models import (Confidence,
                                                      FileMigration,
                                                      StringMigration)
from transifex.native.tools.migrations.review import (
    REVIEW_ACCEPT, FileReviewPolicy, LowConfidenceFileReviewPolicy,
    LowConfidenceStringReviewPolicy, ReviewPolicy, StringReviewPolicy,
    add_line_prefix)


def test_base_class_policy_accepts_all():
    policy = ReviewPolicy()
    assert policy.review_file(_file()) == REVIEW_ACCEPT
    assert policy.review_string(_string(), 1, 1) == REVIEW_ACCEPT


@patch('transifex.native.tools.migrations.review.ReviewPolicy'
       '.prompt_for_string')
@patch('transifex.native.tools.migrations.review.ReviewPolicy'
       '.prompt_for_file')
def test_file_review_policy_prompts_for_file(mock_file_prompt,
                                             mock_string_prompt):
    # This policy prompts for any files
    policy = FileReviewPolicy()
    file_migration = _file()
    policy.review_file(file_migration)
    mock_file_prompt.assert_called_once_with(file_migration)

    # This policy does not prompt for strings
    policy.review_string(_string(Confidence.HIGH),
                         string_cnt=1, strings_total=5)
    policy.review_string(_string(Confidence.LOW),
                         string_cnt=1, strings_total=5)
    assert mock_string_prompt.call_count == 0


@patch('transifex.native.tools.migrations.review.ReviewPolicy'
       '.prompt_for_string')
@patch('transifex.native.tools.migrations.review.ReviewPolicy'
       '.prompt_for_file')
def test_low_file_review_policy_prompts_for_file_with_low_conf_strings(
    mock_file_prompt, mock_string_prompt
):
    # This policy prompts for files that include a string with low confidence
    policy = LowConfidenceFileReviewPolicy()
    file_migration = _file()
    file_migration.add_string(_string(Confidence.HIGH))
    file_migration.add_string(_string(Confidence.HIGH))
    file_migration.add_string(_string(Confidence.LOW))
    policy.review_file(file_migration)
    mock_file_prompt.assert_called_once_with(file_migration)

    # This policy does not prompt for strings
    policy.review_string(_string(Confidence.HIGH),
                         string_cnt=1, strings_total=5)
    policy.review_string(_string(Confidence.LOW),
                         string_cnt=1, strings_total=5)
    assert mock_string_prompt.call_count == 0


@patch('transifex.native.tools.migrations.review.ReviewPolicy'
       '.prompt_for_file')
def test_low_file_review_policy_not_prompts_for_file_with_high_conf_strings(
    mock_file_prompt
):
    # This policy prompts for files that include a string with low confidence
    policy = LowConfidenceFileReviewPolicy()
    file_migration = _file()
    file_migration.add_string(_string(Confidence.HIGH))
    file_migration.add_string(_string(Confidence.HIGH))
    file_migration.add_string(_string(Confidence.HIGH))
    policy.review_file(file_migration)
    assert mock_file_prompt.call_count == 0


@patch('transifex.native.tools.migrations.review.ReviewPolicy'
       '.prompt_for_string')
@patch('transifex.native.tools.migrations.review.ReviewPolicy'
       '.prompt_for_file')
def test_string_review_policy_prompts_for_string(mock_file_prompt,
                                                 mock_string_prompt):
    # This policy prompts for all strings
    policy = StringReviewPolicy()
    string_migration1 = _string(Confidence.HIGH)
    policy.review_string(string_migration1, 5, 10)
    string_migration2 = _string(Confidence.LOW)
    policy.review_string(string_migration2, 15, 20)
    assert mock_string_prompt.call_args_list[0][0] == (
        string_migration1, 5, 10)
    assert mock_string_prompt.call_args_list[1][0] == (
        string_migration2, 15, 20)

    # This policy does not prompt for file reviews
    policy.review_file(_file())
    assert mock_file_prompt.call_count == 0


@patch('transifex.native.tools.migrations.review.ReviewPolicy'
       '.prompt_for_string')
@patch('transifex.native.tools.migrations.review.ReviewPolicy'
       '.prompt_for_file')
def test_low_string_review_policy_prompts_for_low_conf_string_only(
    mock_file_prompt, mock_string_prompt
):
    # This policy prompts for strings that have a low confidence
    policy = LowConfidenceStringReviewPolicy()
    string_migration = _string(Confidence.HIGH)
    policy.review_string(string_migration, 5, 10)
    string_migration = _string(Confidence.LOW)
    policy.review_string(string_migration, 10, 10)
    mock_string_prompt.assert_called_once_with(string_migration, 10, 10)

    # This policy does not prompt for file reviews
    policy.review_file(_file())
    assert mock_file_prompt.call_count == 0


def test_set_comment_format_exception_for_wrong_format():
    # An exception should be raised if the given format does not include {}
    policy = ReviewPolicy()
    with pytest.raises(ValueError):
        policy.set_comment_format('{')


def test_add_line_prefix():
    text = "This\nis\ngood"
    assert add_line_prefix(text, '+ ') == "+ This\n+ is\n+ good"
    assert add_line_prefix(text, '+ ', 99) == \
        "99 + This\n100 + is\n101 + good"
    assert add_line_prefix('', '+ ') == ''
    assert add_line_prefix(None, '+ ') is None


def _string(confidence=Confidence.HIGH):
    """Return a sample StringMigration object for testing."""
    return StringMigration('original', 'new', confidence)


def _file():
    """Return a sample FileMigration object for testing."""
    return FileMigration('filename', 'content')



# -*- coding: utf-8 -*-
import mock
import transifex.native.consts as consts
from django.core.management import call_command
from tests.native.django.test_commands import get_transifex_command
from transifex.native.django.management.commands.transifex import Command
from transifex.native.django.management.common import TranslatableFile
from transifex.native.parsing import SourceString

PYTHON_TEMPLATE = u"""
# -*- coding: utf-8 -*-

{_import}

{call1}(u'{string1}', u'désign1,désign2', param1='1', param2=2, param3=True)
{call2}(
    u'{string2}', u'opération', _comment='comment', _tags='t1,t2', _charlimit=33,
)
"""

HTML_TEMPLATE = u"""
{% load transifex %}

{content}
"""

PATH_FIND_FILES = ('transifex.native.django.management.utils.base.'
                   'CommandMixin._find_files')
PATH_READ_FILE = ('transifex.native.django.management.utils.base.CommandMixin.'
                  '_read_file')
PATH_EXTRACT_STRINGS = ('transifex.native.django.management.utils.push.Push.'
                        '_extract_strings')
PATH_PUSH_STRINGS = ('transifex.native.django.management.utils.push.tx.'
                     'push_source_strings')
PATH_PUSH_STRINGS2 = ('transifex.native.django.management.utils.push.Push.'
                      'push_strings')

PYTHON_FILES = [
    # 1.py
    PYTHON_TEMPLATE.format(
        _import='import transifex.native',
        call1='native.translate',
        call2='native.translate',
        string1=u'Le canapé',
        string2=u'Les données',
    ),
    # 2.py
    PYTHON_TEMPLATE.format(
        _import='import transifex.native as _n',
        call1='_n.translate',
        call2='_n.translate',
        string1=u'Le canapé 2',
        string2=u'Les données 2',
    ),
    # 3.py
    PYTHON_TEMPLATE.format(
        _import='from transifex.native import translate',
        call1='translate',
        call2='translate',
        string1=u'Le canapé 3',
        string2=u'Les données 3',
    ),
]

SOURCE_STRINGS = [
    # 1.py
    SourceString(u'Le canapé', u'désign1,désign2',
                 _occurrences=['r1/1.py:6'],),
    SourceString(
        u'Les données', u'opération', _comment='comment', _tags='t1,t2',
        _charlimit=33, _occurrences=['r1/1.py:7'],
    ),
    # 2.py
    SourceString(u'Le canapé 2', u'désign1,désign2',
                 _occurrences=['r1/dir2/2.py:6'],),
    SourceString(
        u'Les données 2', u'opération', _comment='comment', _tags='t1,t2',
        _charlimit=33, _occurrences=['r1/dir2/2.py:7'],
    ),
    # 3.py
    SourceString(u'Le canapé 3', u'désign1,désign2',
                 _occurrences=['r1/dir3/3.py:6'],),
    SourceString(
        u'Les données 3', u'opération', _comment='comment', _tags='t1,t2',
        _charlimit=33, _occurrences=['r1/dir3/3.py:7'],
    ),
]


@mock.patch(PATH_FIND_FILES)
@mock.patch(PATH_READ_FILE)
def test_python_parsing_raises_unicode_error(mock_read, mock_find_files):
    o = b'\x00\x00'
    mock_read.side_effect = UnicodeDecodeError(
        'funnycodec', o, 1, 2, 'This is just a fake reason!')
    mock_find_files.return_value = [
        TranslatableFile('dir1', '1.py', 'locdir1'),
    ]

    command = get_transifex_command()
    call_command(command, 'push')
    # command.string_collection.strings is like: {<key>: <SourceString>}
    found = command.subcommands['push'].string_collection.strings.values()
    assert set(found) == set([])


@mock.patch(PATH_FIND_FILES)
@mock.patch(PATH_EXTRACT_STRINGS)
def test_python_parsing_no_strings(mock_extract, mock_find_files):
    mock_extract.return_value = []
    mock_find_files.return_value = [
        TranslatableFile('dir1', '1.py', 'locdir1'),
        TranslatableFile('dir1/dir2', '2.py', 'locdir1'),
        TranslatableFile('dir1/dir3', '3.py', 'locdir1'),
    ]

    command = get_transifex_command()
    call_command(command, 'push')
    # command.string_collection.strings is like: {<key>: <SourceString>}
    found = command.subcommands['push'].string_collection.strings.values()
    assert set(found) == set([])


@mock.patch(PATH_PUSH_STRINGS)
@mock.patch(PATH_READ_FILE)
@mock.patch(PATH_FIND_FILES)
def test_python_parsing_push_exception(mock_find_files, mock_read, mock_push_strings):
    mock_push_strings.return_value = 500, "content_to_trigger_exception"
    mock_find_files.return_value = [
        TranslatableFile('dir1', '1.py', 'locdir1'),
        TranslatableFile('dir1/dir2', '2.py', 'locdir1'),
        TranslatableFile('dir1/dir3', '3.py', 'locdir1'),
    ]
    mock_read.side_effect = PYTHON_FILES

    expected = SOURCE_STRINGS
    run_and_compare(expected)


@mock.patch(PATH_PUSH_STRINGS)
@mock.patch(PATH_READ_FILE)
@mock.patch(PATH_FIND_FILES)
def test_python_parsing_success(mock_find_files, mock_read, mock_push_strings):
    mock_push_strings.return_value = 200, {'doesnt': 'matter'}
    mock_find_files.return_value = [
        TranslatableFile('dir1', '1.py', 'locdir1'),
        TranslatableFile('dir1/dir2', '2.py', 'locdir1'),
        TranslatableFile('dir1/dir3', '3.py', 'locdir1'),
    ]
    mock_read.side_effect = PYTHON_FILES

    expected = SOURCE_STRINGS
    run_and_compare(expected)


@mock.patch(PATH_PUSH_STRINGS)
@mock.patch(PATH_READ_FILE)
@mock.patch(PATH_FIND_FILES)
def test_append_tags(mock_find_files, mock_read, mock_push_strings):
    """Test the functionality of the --append_tags option.

    The new tags should be added to the existing ones of each string.
    """
    mock_push_strings.return_value = 200, {'doesnt': 'matter'}
    mock_find_files.return_value = [
        TranslatableFile('dir1', '1.py', 'locdir1'),
    ]
    mock_read.side_effect = PYTHON_FILES

    expected = [
        clone_string(SOURCE_STRINGS[0], new_tags=['extra1', 'extra2']),
        clone_string(SOURCE_STRINGS[1], new_tags=[
                     't1', 't2', 'extra1', 'extra2']),
    ]
    run_and_compare(expected, append_tags=u'extra1,extra2')


@mock.patch(PATH_PUSH_STRINGS)
@mock.patch(PATH_READ_FILE)
@mock.patch(PATH_FIND_FILES)
def test_with_tags_only(mock_find_files, mock_read, mock_push_strings):
    mock_push_strings.return_value = 200, {'doesnt': 'matter'}
    mock_find_files.return_value = [
        TranslatableFile('dir1', '1.py', 'locdir1'),
        TranslatableFile('dir1/dir2', '2.py', 'locdir1'),
        TranslatableFile('dir1/dir3', '3.py', 'locdir1'),
    ]
    mock_read.side_effect = PYTHON_FILES

    expected = [
        clone_string(SOURCE_STRINGS[1]),
        clone_string(SOURCE_STRINGS[3]),
        clone_string(SOURCE_STRINGS[5]),
    ]
    run_and_compare(expected, with_tags_only='t1,t2')


@mock.patch(PATH_PUSH_STRINGS)
@mock.patch(PATH_READ_FILE)
@mock.patch(PATH_FIND_FILES)
def test_without_tags_only(mock_find_files, mock_read, mock_push_strings):
    mock_push_strings.return_value = 200, {'doesnt': 'matter'}
    mock_find_files.return_value = [
        TranslatableFile('dir1', '1.py', 'locdir1'),
        TranslatableFile('dir1/dir2', '2.py', 'locdir1'),
        TranslatableFile('dir1/dir3', '3.py', 'locdir1'),
    ]
    mock_read.side_effect = PYTHON_FILES

    expected = [
        clone_string(SOURCE_STRINGS[0]),
        clone_string(SOURCE_STRINGS[2]),
        clone_string(SOURCE_STRINGS[4]),
    ]
    run_and_compare(expected, without_tags_only='t1,t2')


@mock.patch(PATH_PUSH_STRINGS)
@mock.patch(PATH_READ_FILE)
@mock.patch(PATH_FIND_FILES)
def test_all_tags_options(mock_find_files, mock_read, mock_push_strings):
    mock_push_strings.return_value = 200, {'doesnt': 'matter'}
    mock_find_files.return_value = [
        TranslatableFile('dir1', '1.py', 'locdir1'),
        TranslatableFile('dir1/dir2', '2.py', 'locdir1'),
        TranslatableFile('dir1/dir3', '3.py', 'locdir1'),
    ]
    mock_read.side_effect = PYTHON_FILES

    expected = [
        clone_string(SOURCE_STRINGS[0], new_tags=['extra1', 'extra2']),
        clone_string(SOURCE_STRINGS[2], new_tags=['extra1', 'extra2']),
        clone_string(SOURCE_STRINGS[4], new_tags=['extra1', 'extra2']),
    ]
    run_and_compare(
        expected,
        append_tags='extra1,extra2',
        with_tags_only='extra1',
        without_tags_only='t1,t2',
    )


@mock.patch(PATH_PUSH_STRINGS2)
@mock.patch(PATH_READ_FILE)
@mock.patch(PATH_FIND_FILES)
def test_template_parsing(mock_find_files, mock_read, mock_push_strings):
    mock_find_files.return_value = [
        TranslatableFile('dir1/dir2', '1.html', 'locdir1'),
        TranslatableFile('dir4/dir5', '1.txt', 'locdir1'),
    ]
    mock_read.side_effect = [
        # 1.html
        HTML_TEMPLATE.replace(
            '{content}',
            u'<p>{% t "<b>Strong</b> {a}" a="A" _context="c1,c2" '
            u'_tags="t1,t2" _comment="comment1" _charlimit=22 %}</p>\n'

            u'<p>{% ut "παράδειγμα {b}" b="B" _context="c1,c2" '
            u'_tags="t1,t2" _comment="comment2" _charlimit=33 %}</p>'
        ),
        # 1.txt
        HTML_TEMPLATE.replace(
            '{content}',
            u'{% t _context="c1,c2" _tags="t1,t2" _comment="co1" _charlimit=22 %}\n'
            u'This is a short string\n'
            u'{% endt %}\n'

            u'{% t _context="c1,c2" _tags="t1,t2" _comment="co2" _charlimit=33 %}\n'
            u'This is not a shorter string\n'
            u'{% endt %}'
        ),
    ]

    expected = [
        # 1.html
        SourceString(
            u'<b>Strong</b> {a}', 'c1,c2', _tags='t1,t2', _comment="comment1",
            _charlimit=22, _occurrences=['r1/dir2/1.html:4'],
        ),
        SourceString(
            u'παράδειγμα {b}', 'c1,c2', _tags='t1,t2', _comment="comment2",
            _charlimit=33, _occurrences=['r1/dir2/1.html:5'],
        ),

        # 1.txt
        SourceString(
            u'\nThis is a short string\n', 'c1,c2', _tags='t1,t2',
            _comment="co1", _charlimit=22, _occurrences=['r4/dir5/1.txt:4'],
        ),
        SourceString(
            u'\nThis is not a shorter string\n', 'c1,c2', _tags='t1,t2',
            _comment="co2", _charlimit=33, _occurrences=['r4/dir5/1.txt:7'],
        ),
    ]
    run_and_compare(expected)


@mock.patch(PATH_PUSH_STRINGS2)
@mock.patch(PATH_READ_FILE)
@mock.patch(PATH_FIND_FILES)
def test_no_detection_for_non_transifex(mock_find_files, mock_read, mock_push_strings):
    """No strings should be detected if a format other than Transifex Native
    is used in Python files and templates.
    """
    mock_find_files.return_value = [
        # 2 files with valid extension but no translatable content
        TranslatableFile('dir4/dir5', 'empty.py', 'locdir1'),
        TranslatableFile('dir4/dir5', 'empty.txt', 'locdir1'),
    ]
    mock_read.side_effect = [
        # empty.py - shouldn't detect any strings as non-transifex
        PYTHON_TEMPLATE.format(
            _import='from django.utils.translation import ugettext_lazy as _',
            call1='_',
            call2='_',
            string1=u'A Django string',
            string2=u'Another Django string',
        ),
        # empty.txt - shouldn't detect any strings as non-transifex
        (
            u'{% load i18n %}\n'
            u'{% trans "A Django string %}\n'
            u'{% blocktrans %}Another Django string{% endblocktrans %}',
        ),
    ]
    command = get_transifex_command()
    call_command(command, 'push', domain='djangojs')
    # command.string_collection.strings is like: {<key>: <SourceString>}
    found = command.subcommands['push'].string_collection.strings.values()
    assert set(found) == set([])


@mock.patch(PATH_PUSH_STRINGS)
@mock.patch(PATH_READ_FILE)
@mock.patch(PATH_FIND_FILES)
def test_dry_run(mock_find_files, mock_read, mock_push_strings):
    mock_push_strings.return_value = 200, {'doesnt': 'matter'}
    mock_find_files.return_value = [
        TranslatableFile('dir1', '1.py', 'locdir1'),
    ]
    mock_read.side_effect = PYTHON_FILES

    command = get_transifex_command()
    call_command(command, 'push', dry_run=1)
    assert not mock_push_strings.called


def clone_string(source_string, new_tags=None):
    """Create a new SourceString instance, identical to the one given,
    by optionally replacing its tags with the ones given.

    :param SourceString source_string: the string to clone
    :param list new_tags: a list of strings to use as the tags
        for the new string
    :rtype: SourceString
    :return: a new SourceString instance
    """
    cloned_string = SourceString(
        source_string.string,
        _context=(
            ','.join(source_string.context)
            if source_string.context else ''
        ),
        **source_string.meta
    )
    if new_tags is not None:
        cloned_string.meta[consts.KEY_TAGS] = new_tags
    return cloned_string


def run_and_compare(expected, **kwargs):
    """Run the command and compare the detected strings with the expected ones.

    Any given kwarg is passed as an argument to the command.
    For example `run_and_compare([...], append_tags=u'extra1,extra2')
    is equivalent to running with `--append_tags=extra1,extra2`.

    :param list expected: a list of SourceString objects
    """
    command = get_transifex_command()
    call_command(command, 'push', **kwargs)
    # command.string_collection.strings is like: {<key>: <SourceString>}
    found = command.subcommands['push'].string_collection.strings.values()
    assert set(found) == set(expected)



try:
    from django.template.base import (TOKEN_BLOCK, TOKEN_COMMENT,
                                      TOKEN_TEXT, TOKEN_VAR)
except ImportError:
    from django.template.base import TokenType
    TOKEN_BLOCK = TokenType.BLOCK
    TOKEN_COMMENT = TokenType.COMMENT
    TOKEN_TEXT = TokenType.TEXT
    TOKEN_VAR = TokenType.VAR



from operator import itemgetter

import pytest
import responses
from mock import patch
from transifex.native.cds import CDSHandler
from transifex.native.parsing import SourceString


class TestCDSHandler(object):

    def _lang_lists_equal(self, list_1, list_2):
        try:
            sorted_1, sorted_2 = [sorted(language, key=itemgetter('code'))
                                  for language in (list_1, list_2)]
            pairs = zip(sorted_1, sorted_2)
            difference = any(x != y for x, y in pairs)
        except Exception:
            difference = True
        return not difference

    @responses.activate
    @patch('transifex.native.cds.logger')
    def test_fetch_languages(self, patched_logger):
        cds_host = 'https://some.host'
        cds_handler = CDSHandler(
            ['el', 'en'],
            'some_token',
            host=cds_host
        )

        # correct response
        responses.add(
            responses.GET, cds_host + '/languages',
            json={
                "data": [
                    {
                        "code": "el",
                    },
                    {
                        "code": "en",
                    },
                ],
                "meta": {
                    "some_key": "some_value"
                }
            }, status=200
        )

        languages_response = cds_handler.fetch_languages()
        assert self._lang_lists_equal(
            languages_response,
            [{'code': 'el'}, {'code': 'en'}]
        )
        assert patched_logger.error.call_count == 0
        responses.reset()

        # wrong payload structure
        responses.add(
            responses.GET, cds_host + '/languages',
            json={
                "wrong_key": [
                    {
                        "code": "el",
                    },
                    {
                        "code": "en",
                    },
                ],
                "meta": {
                    "some_key": "some_value"
                }
            }, status=200
        )

        assert cds_handler.fetch_languages() == []
        patched_logger.error.assert_called_with(
            'Error retrieving languages from CDS: Malformed response'
        )
        responses.reset()

        # bad request
        responses.add(
            responses.GET, cds_host + '/languages',
            json={
                "data": [
                    {
                        "code": "el",
                    },
                    {
                        "code": "en",
                    },
                ],
                "meta": {
                    "some_key": "some_value"
                }
            }, status=400
        )

        assert cds_handler.fetch_languages() == []
        patched_logger.error.assert_called_with(
            'Error retrieving languages from CDS: UnknownError (`400 Client '
            'Error: Bad Request for url: https://some.host/languages`)'
        )
        responses.reset()

        # unauthorized
        responses.add(
            responses.GET, cds_host + '/languages',
            json={
                "data": [
                    {
                        "code": "el",
                    },
                    {
                        "code": "en",
                    },
                ],
                "meta": {
                    "some_key": "some_value"
                }
            }, status=403
        )

        assert cds_handler.fetch_languages() == []
        patched_logger.error.assert_called_with(
            'Error retrieving languages from CDS: UnknownError (`403 Client '
            'Error: Forbidden for url: https://some.host/languages`)'
        )
        responses.reset()

        # connection error
        assert cds_handler.fetch_languages() == []
        patched_logger.error.assert_called_with(
            'Error retrieving languages from CDS: ConnectionError'
        )
        responses.reset()

    @responses.activate
    @patch('transifex.native.cds.logger')
    def test_fetch_translations(self, patched_logger):
        cds_host = 'https://some.host'
        cds_handler = CDSHandler(
            ['el', 'en', 'fr'],
            'some_token',
            host=cds_host
        )

        # add response for languages
        responses.add(
            responses.GET, cds_host + '/languages',
            json={
                "data": [
                    {
                        "code": "el",
                    },
                    {
                        "code": "en",
                    },
                    {
                        "code": "fr",
                    },
                ],
                "meta": {
                    "some_key": "some_value"
                }
            }, status=200
        )

        # add response for translations
        responses.add(
            responses.GET, cds_host + '/content/el',
            json={
                'data': {
                    'key1': {
                        'string': 'key1_el'
                    },
                    'key2': {
                        'string': 'key2_el'
                    },
                },
                'meta': {
                    "some_key": "some_value"
                }
            }, status=200
        )

        responses.add(
            responses.GET, cds_host + '/content/en',
            json={
                'data': {
                    'key1': {
                        'string': 'key1_en'
                    },
                    'key2': {
                        'string': 'key2_en'
                    },
                },
                'meta': {
                }
            }, status=200
        )

        # add response bad status response for a language here
        responses.add(
            responses.GET, cds_host + '/content/fr', status=404
        )

        resp = cds_handler.fetch_translations()
        assert resp == {
            'el': (True, {
                'key1': {
                    'string': 'key1_el'
                },
                'key2': {
                    'string': 'key2_el'
                },
            }),
            'en': (True, {
                'key1': {
                    'string': 'key1_en'
                },
                'key2': {
                    'string': 'key2_en'
                },
            }),
            'fr': (False, {})  # that is due to the error status in response
        }

        responses.reset()

        # test fetch_languages fails with connection error
        responses.add(responses.GET, cds_host + '/languages', status=500)
        resp = cds_handler.fetch_translations()
        assert resp == {}

        patched_logger.error.assert_called_with(
            'Error retrieving languages from CDS: UnknownError '
            '(`500 Server Error: Internal Server Error for url: '
            'https://some.host/languages`)'
        )
        responses.reset()
        patched_logger.reset_mock()

        # test language code
        responses.add(
            responses.GET, cds_host + '/content/el',
            json={
                'data': {
                    'key1': {
                        'string': 'key1_el'
                    },
                    'key2': {
                        'string': 'key2_el'
                    },
                },
                'meta': {
                    "some_key": "some_value"
                }
            }, status=200
        )

        resp = cds_handler.fetch_translations(language_code='el')
        assert resp == {
            'el': (True, {
                'key1': {
                    'string': 'key1_el'
                },
                'key2': {
                    'string': 'key2_el'
                },
            })
        }
        responses.reset()
        assert patched_logger.error.call_count == 0

        # test connection_error
        resp = cds_handler.fetch_translations(language_code='el')
        patched_logger.error.assert_called_with(
            'Error retrieving translations from CDS: ConnectionError'
        )
        assert resp == {'el': (False, {})}

    @responses.activate
    @patch('transifex.native.cds.logger')
    def test_fetch_translations_etags_management(self, patched_logger):

        cds_host = 'https://some.host'
        cds_handler = CDSHandler(
            ['el', 'en'],
            'some_token',
            host=cds_host
        )

        # add response for languages
        responses.add(
            responses.GET, cds_host + '/languages',
            json={
                "data": [
                    {
                        "code": "el",
                    },
                    {
                        "code": "en",
                    },
                ],
                "meta": {
                    "some_key": "some_value"
                }
            }, status=200
        )

        # add response for translations
        responses.add(
            responses.GET, cds_host + '/content/el',
            json={
                'data': {
                    'key1': {
                        'string': 'key1_el'
                    },
                    'key2': {
                        'string': 'key2_el'
                    },
                },
                'meta': {
                    "some_key": "some_value"
                }
            },
            status=200,
            headers={'ETag': 'some_unique_tag_is_here'}
        )

        responses.add(
            responses.GET, cds_host + '/content/en',
            # whatever, we don't care about the content of json repsone atm.
            json={},
            status=304
        )

        resp = cds_handler.fetch_translations()
        assert resp == {
            'el': (True, {
                'key1': {
                    'string': 'key1_el'
                },
                'key2': {
                    'string': 'key2_el'
                },
            }),
            'en': (False, {})
        }
        assert cds_handler.etags.get('el') == 'some_unique_tag_is_here'

    def test_push_source_strings_no_secret(self):
        cds_handler = CDSHandler(
            ['el', 'en'],
            'some_token',
        )
        with pytest.raises(Exception):
            cds_handler.push_source_strings([], False)

    @responses.activate
    @patch('transifex.native.cds.logger')
    def test_push_source_strings(self, patched_logger):
        cds_host = 'https://some.host'
        cds_handler = CDSHandler(
            ['el', 'en'],
            'some_token',
            secret='some_secret',
            host=cds_host
        )

        # test push no correct
        responses.add(
            responses.POST, cds_host + '/content/',
            status=200, json={'data': []}
        )

        cds_handler.push_source_strings([], False)
        assert patched_logger.error.call_count == 0

        # test push with content
        responses.add(
            responses.POST, cds_host + '/content/',
            status=200, json={'data': []}
        )

        source_string = SourceString('some_string')
        cds_handler.push_source_strings([source_string], False)
        assert patched_logger.error.call_count == 0
        responses.reset()

        # test wrong data format
        responses.add(
            responses.POST, cds_host + '/content/',
            status=422, json={
                "status": 422,
                "message": "Invalid Payload",
                "details": [
                    {
                        "message": "\"string\" is required",
                        "path": [
                            "some_key1",
                            "string"
                        ],
                        "type": "any.required",
                        "context": {
                            "key": "string",
                            "label": "string"
                        }
                    }
                ]
            }
        )
        # we don't care about the payload this time, just want to
        # see how the service handles the errors
        cds_handler.push_source_strings([], False)
        # The actual error message differs between Python 2 and Python 3
        messages = [
            'Error pushing source strings to CDS: UnknownError '
            '(`422 Client Error: {err} for url: '
            'https://some.host/content/`)'.format(err=x)
            for x in ('Unprocessable Entity', 'None')
        ]
        assert patched_logger.error.call_args[0][0] in messages

    def test_get_headers(self):
        cds_host = 'https://some.host'
        cds_handler = CDSHandler(
            ['el', 'en'],
            'some_token',
            secret='some_secret',
            host=cds_host
        )
        assert cds_handler._get_headers() == {
            'Authorization': 'Bearer some_token',
            'Accept-Encoding': 'gzip',
            'X-NATIVE-SDK': 'python',
        }

        assert cds_handler._get_headers(use_secret=True) == {
            'Authorization': 'Bearer some_token:some_secret',
            'Accept-Encoding': 'gzip',
            'X-NATIVE-SDK': 'python',
        }

        headers = cds_handler._get_headers(
            use_secret=True, etag='something')
        assert headers == {
            'Authorization': 'Bearer some_token:some_secret',
            'Accept-Encoding': 'gzip',
            'X-NATIVE-SDK': 'python',
            'If-None-Match': 'something',
        }

    @responses.activate
    def test_retry_fetch_languages(self):
        cds_host = 'https://some.host'
        cds_handler = CDSHandler(
            ['el', 'en'],
            'some_token',
            host=cds_host,
        )
        responses.add(responses.GET, cds_host + '/languages', status=500)
        responses.add(responses.GET, cds_host + '/languages', status=202)
        responses.add(responses.GET, cds_host + '/languages',
                      json={'data': [{'code': "el"},
                                     {'code': "en"}],
                            'meta': {'some_key': "some_value"}},
                      status=200)
        languages_response = cds_handler.fetch_languages()
        assert self._lang_lists_equal(
            languages_response,
            [{'code': 'el'}, {'code': 'en'}]
        )

    @responses.activate
    def test_retry_fetch_translations(self):
        cds_host = 'https://some.host'
        cds_handler = CDSHandler(
            ['el', 'en'],
            'some_token',
            host=cds_host,
        )
        responses.add(responses.GET, cds_host + '/content/el', status=500)
        responses.add(responses.GET, cds_host + '/content/el', status=202)
        responses.add(responses.GET,
                      cds_host + '/content/el',
                      json={'data': {'source': {'string': "translation"}}},
                      status=200)
        translations = cds_handler.fetch_translations('el')
        assert (translations ==
                {'el': (True, {'source': {'string': "translation"}})})

    def test_invalidate_no_secret(self):
        cds_handler = CDSHandler(
            ['el', 'en'],
            'some_token',
        )
        with pytest.raises(Exception):
            cds_handler.invalidate_cache(False)

    @responses.activate
    @patch('transifex.native.cds.logger')
    def test_invalidate(self, patched_logger):
        cds_host = 'https://some.host'
        cds_handler = CDSHandler(
            ['el', 'en'],
            'some_token',
            secret='some_secret',
            host=cds_host
        )

        # test invalidate
        responses.add(
            responses.POST, cds_host + '/invalidate',
            status=200, json={'count': 5}
        )

        cds_handler.invalidate_cache(False)
        assert patched_logger.error.call_count == 0

        # test purge
        responses.add(
            responses.POST, cds_host + '/purge',
            status=200, json={'count': 5}
        )

        cds_handler.invalidate_cache(True)
        assert patched_logger.error.call_count == 0
        responses.reset()

        # test response error
        responses.add(
            responses.POST, cds_host + '/invalidate',
            status=422, json={
                "status": 422,
            }
        )
        # we don't care about the payload this time, just want to
        # see how the service handles the errors
        cds_handler.invalidate_cache(False)
        # The actual error message differs between Python 2 and Python 3
        messages = [
            'Error invalidating CDS: UnknownError '
            '(`422 Client Error: {err} for url: '
            'https://some.host/invalidate`)'.format(err=x)
            for x in ('Unprocessable Entity', 'None')
        ]
        assert patched_logger.error.call_args[0][0] in messages



# -*- coding: utf-8 -*-
"""This module contains everything related to the functionality
that marks a migration to Transifex Native code for proofreading.

When a file or a string in a file is marked for proofreading, a special
string is added in the file, so that users can grep for it and
make sure the migration is correct.
"""

from __future__ import unicode_literals

from transifex.common.console import Color
from transifex.native.tools.migrations.models import Confidence

MARK_PROOFREAD_STRING = 'Transifex Native: PROOFREAD_STRING'
MARK_PROOFREAD_FILE = 'Transifex Native: PROOFREAD_FILE'
MARK_ARGUMENT_FIXME = '__txnative_fixme'


def mark_string(string_migration, comment_format, mark):
    """Mark a string migration for proofreading.

    Adds a comment in the beginning of this string.

    :param StringMigration string_migration: the string migration object
        to update
    :param unicode comment_format: the format of the comment,
        compatible with the file type of the migration
    :param str mark: the string to add as a comment
    """
    string_migration.update(
        '', comment_format.format(mark),
        append=False,
    )


class MarkPolicy(object):
    """Determines if a migrated file or a migrated string will be marked
    for proofread.
    """
    name = None

    def __init__(self):
        # This determines how review comments will appear
        # inside the migrated file. It is dynamic because different files
        # support different comment formats, e.g.
        # - '# Comment' in Python
        # - '<!-- Comment -->' in HTML
        self._comment_format = '{}'

    def set_comment_format(self, comment_format):
        """Define the comment format to use when a policy adds
        comments to the migrated file.

        Example:
        >>> set_comment_format('# {}')

        :param unicode comment_format: the format to use; must include '{}'
        :raise ValueError: if `comment_format` does not include '{}'
        """
        if '{}' not in comment_format:
            raise ValueError('The provided `comment_format` must include {}')
        self._comment_format = comment_format

    def should_mark_strings(self):
        """Returns whether or not a policy object wants to mark individual
        strings.

        Although policy subclasses can only mark *some* strings, and not all,
        this method is important for optimization, so that the objects that
        use a policy object know if they need to feed each string migration
        to the mark policy or not.
        """
        return False

    def mark_file(self, file_migration):
        """The base class does nothing.

        :param FileMigration file_migration: the migration object
        :return: True if the file was marked, False otherwise
        :rtype: bool
        """
        return False

    def mark_string(self, string_migration):
        """The base class does nothing.

        :param StringMigration string_migration: the migration object
        :return: True if the string was marked, False otherwise
        :rtype: bool
        """
        return False


class NoopMarkPolicy(MarkPolicy):
    """Does not mark anything."""

    name = 'none'


class MarkLowConfidenceFilesPolicy(MarkPolicy):
    """Marks all files that have at least one string with low confidence."""

    name = 'file-low'

    def mark_file(self, file_migration):
        if not file_migration.low_confidence_strings:
            return False

        first_string_migration = file_migration.strings[0]
        if MARK_PROOFREAD_FILE in first_string_migration.new:
            return False

        mark_string(
            first_string_migration,
            self._comment_format,
            MARK_PROOFREAD_FILE,
        )
        Color.echo('📝 File automatically marked for proofreading')
        return True


class MarkLowConfidenceStringsPolicy(MarkPolicy):
    """Marks all string that have low confidence."""

    name = 'string-low'

    def should_mark_strings(self):
        return True

    def mark_string(self, string_migration):
        if string_migration.confidence != Confidence.LOW:
            return False

        if MARK_PROOFREAD_STRING in string_migration.new:
            return False

        mark_string(
            string_migration,
            self._comment_format,
            MARK_PROOFREAD_STRING,
        )
        return True


def create_mark_policy(policy_id):
    """Create the mark policy object that corresponds to the given ID.

    :param str policy_id: the ID of the policy to create, as defined
        in the MarkPolicy subclasses
    :return: an instance of a subclass of MarkPolicy
    :rtype: MarkPolicy
    """
    policy_classes = {
        x.name: x
        for x in [
            NoopMarkPolicy, MarkLowConfidenceFilesPolicy,
            MarkLowConfidenceStringsPolicy,
        ]
    }
    try:
        _class = policy_classes[policy_id.lower()]
        return _class()
    except KeyError:
        raise AttributeError('Invalid mark policy ID={}'.format(policy_id))



from transifex.native.django.management.commands.transifex import Command


def get_transifex_command():
    """Return an instance of the Transifex Django command
    that works in all Django versions.

    After Django 2.0+, attempting to run
    `call_command(command, 'subcommand', option=value)`
    with `option` not being defined on the main command results
    to a TypeError, e.g. Unknown option(s) for transifex command: dry_run

    This is solved by adding all possible options to the command's
    `stealth_options` attribute.
    """
    command = Command()
    command.stealth_options = (
        # Migrate
        'files',
        'path',
        'text',
        'save_policy',
        'review_policy',
        'mark_policy',
        'verbose',

        # Push
        'extension',
        'append_tags',
        'with_tags_only',
        'without_tags_only',
        'dry_run',
        'symlinks',

        # Invalidate
        'purge',
    )
    return command



from transifex.common.utils import parse_plurals


def test_simple():
    assert (False, {5: "hello world"}) == parse_plurals("hello world")

    assert (True, {1: 'ONE', 5: 'OTHER'}) == \
        parse_plurals("{cnt, plural, one {ONE} other {OTHER}}")

    assert (True, {1: 'ONE', 5: 'OTHER'}) == \
        parse_plurals("{cnt, plural, one {ONE} =5 {OTHER}}")

    assert (True, {1: 'ONE', 5: 'OTHER'}) == \
        parse_plurals("{cnt, plural, =1 {ONE} other {OTHER}}")

    assert (True, {1: 'ONE', 5: 'OTHER'}) == \
        parse_plurals("{cnt, plural, =1 {ONE} =5 {OTHER}}")


def test_almost_plural():
    assert (False, {5: "{cnt, plural, one {ONE} other {OTHER}"}) ==\
        parse_plurals("{cnt, plural, one {ONE} other {OTHER}")
    assert (False, {5: "{cnt, plurall, one {ONE} other {OTHER}}"}) ==\
        parse_plurals("{cnt, plurall, one {ONE} other {OTHER}}")
    assert (False, {5: "{cnt, plural, onee {ONE} other {OTHER}}"}) ==\
        parse_plurals("{cnt, plural, onee {ONE} other {OTHER}}")
    assert (False, {5: "{cnt, plural, =7 {ONE} other {OTHER}}"}) ==\
        parse_plurals("{cnt, plural, =7 {ONE} other {OTHER}}")
    assert (False, {5: "{cnt, plural, one {ONE}, other {OTHER}}"}) ==\
        parse_plurals("{cnt, plural, one {ONE}, other {OTHER}}")
    assert (False, {5: "{cnt, plural, one {ONE} many {OTHER}}"}) ==\
        parse_plurals("{cnt, plural, one {ONE} many {OTHER}}")
    assert (False, {5: "{cnt, plural, one {ONE}}"}) ==\
        parse_plurals("{cnt, plural, one {ONE}}")


def test_escapes():
    assert (True, {1: "O '' NE", 5: 'OTHER'}) == \
        parse_plurals("{cnt, plural, one {O '' NE} other {OTHER}}")
    assert (True, {1: "O '{ NE'", 5: 'OTHER'}) == \
        parse_plurals("{cnt, plural, one {O '{ NE'} other {OTHER}}")



from __future__ import absolute_import, print_function, unicode_literals

import itertools
import json
import re
import sys

from django.template import Context, Template
from django.utils import translation
from transifex.common.console import Color
from transifex.native.django.management.utils.base import CommandMixin

try:
    raw_input
except NameError:
    pass
else:
    input = raw_input  # pragma no cover  # noqa


def fancy_input(text, *choices):
    """ Multiple choice input

        Given arguments ('Which tag do you want to use?',
                         ('t', 't', '{% t ... %}'),
                         ('ut', 'ut', '{% ut ... %}'))

        It will render:

            ===> Which tag do you want to use?
            .... 1. t                 : {% t ... %}
            .... 2. ut                : {% ut ... %}
            .... [examples: "1", "1 3"; empty input for all choices]

        The return value will be a list of "choices", ie the first items in the
        3-tuples.
    """

    print()
    Color.echo("[yel]===>[end] {}".format(text))
    for i, (_, display, example) in enumerate(choices, 1):
        line = "[yel]....[end] [warn]{}[end]. {:18}".format(i, display)
        if example:
            line += ": [cyan]{}[end]".format(example)
        Color.echo(line)

    while True:
        Color.echo('.... [examples: "[warn]1[end]", "[warn]1 3[end]"; '
                   '[warn]empty input[end] for all choices]')
        answer = input("===> ")
        if answer.strip() == "":
            return [choice for choice, _, _ in choices]
        try:
            answer = [int(choice) - 1 for choice in answer.split()]
        except Exception:
            pass
        else:
            if all((0 <= choice < len(choices) for choice in answer)):
                return [choices[choice][0] for choice in answer]
        print("Invalid answer, please try again")


def make_tests(tag_names, sources, source_filters, params, param_filters,
               asvars, asvar_filters, blocks, context_values):
    """ Use `itertools.product` to create a set of tests. Each argument should
        be a list with at least one item (otherwise `product` will return an
        empty result). It delegates each test creation to the `make_test`
        method.
    """

    results = []
    for (tag_name, source, source_filter, param, param_filter, asvar,
            asvar_filter, block,
            context_value) in itertools.product(
                tag_names, sources, source_filters, params, param_filters,
                asvars, asvar_filters, blocks, context_values
    ):
        try:
            test = make_test(tag_name, source, source_filter, param,
                             param_filter, asvar, asvar_filter, block,
                             context_value)
        except TypeError:
            pass
        else:
            results.append(test)

    # Remove duplicates
    results = [(template, json.dumps(context))
               for template, context in results]
    results = sorted(set(results))
    results = [(template, json.loads(context))
               for template, context in results]
    return results


def make_test(tag_name, source=None, source_filter=None, param=None,
              param_filter=None, asvar=None, asvar_filter=None, block=None,
              context_value=None):
    """ Compose a template and context for testing based on the arguments.  """

    if not source and not block:
        raise TypeError("At least one of 'source' or 'block' must be "
                        "specified")

    result = ['{% ', tag_name, ' ']
    context_vars = []
    if source:
        if source[0] not in ('"', "'"):
            context_vars.append(source)
        for match in re.findall(r'\{[\w_]+\}', source):
            context_vars.append(match[1:-1])
        result.append(source)
    if source_filter:
        result.extend(['|', source_filter])
    if param:
        context_vars.append(param)
        result.extend([' ', param, '=', param])
        if param_filter:
            result.extend(['|', param_filter])
    if asvar:
        result.extend([' as ', asvar])
    result.append(' %}')
    if not source and block:
        for match in re.findall(r'\{[\w_]+\}', block):
            context_vars.append(match[1:-1])
        result.extend([block, '{% end', tag_name, ' %}'])
    if asvar:
        result.extend(['{{ ', asvar])
        if asvar_filter:
            result.extend(['|', asvar_filter])
        result.extend([' }}'])

    context = {}
    if context_value is not None:
        context = {context_var: context_value for context_var in context_vars}

    return ''.join(result), context


def test(template_str, context_dict=None, autoescape=True, i=''):
    """ Use the django templating engine to run a test.

        Arguments:

        :param template_str: The template to render
        :param context_dict: The context to render the template against
        :param autoescape:   Pretend the django templating engine was setup
                             with autoescape or not (in most real use-cases, it
                             will have been set up with autoescape=True)
        :param i:            Prepend the output with this in order to help
                             distinguish tests when multiple are run

        Information about (auto)escaping in django:
        https://docs.djangoproject.com/en/3.0/ref/templates/language/#automatic-html-escaping  # noqa
    """

    if context_dict is None:
        context_dict = {}
    context = Context(dict(context_dict), autoescape=autoescape)
    template = ('{% load transifex %}' + template_str)
    try:
        result = Template(template).render(context)
    except Exception:
        print(template_str, context_dict, autoescape)
        raise
    Color.echo("[warn]{i:4}[end]. [cyan]Template[end]:    {template}".
               format(i=i, template=template_str))
    Color.echo("      [cyan]Context[end]:     {context}".
               format(context=context_dict))
    Color.echo("      [cyan]Autoescape[end]:  {autoescape}".
               format(autoescape=autoescape))
    Color.echo("      [cyan]Result[end]:      [green]{result}[end]".
               format(result=result))
    print()


class TryTemplatetag(CommandMixin):
    """ Command to help users try cornercases when using our templatetag,
        especially related to escaping. See the help attribute of main parser
        in `add_arguments` for details.
    """

    def add_arguments(self, subparsers):
        parser = subparsers.add_parser(
            'try-templatetag',
            help="""
                Test the transifex templatetag against multiple variations of
                test cases.

                We recommend you generate your test cases with the
                `--interactive` option. At the end, it will provide you with
                the detailed invocation you would need in order to repeat the
                same tests, which you can edit to fine-tune your testing.
            """
        )

        parser.add_argument('-i', '--interactive', action="store_true",
                            dest="interactive",
                            help="Interactive dialog for setting up tests")
        parser.add_argument('--tag-names', default='', dest="tag_names")
        parser.add_argument('--sources', default='', dest="sources")
        parser.add_argument('--source-filters', default='',
                            dest="source_filters")
        parser.add_argument('--params', default='', dest="params")
        parser.add_argument('--param-filters', default='',
                            dest="param_filters")
        parser.add_argument('--asvars', default='', dest="asvars")
        parser.add_argument('--asvar-filters', default='',
                            dest="asvar_filters")
        parser.add_argument('--blocks', default='', dest="blocks")
        parser.add_argument('--context-values', default='',
                            dest="context_values")
        parser.add_argument('--autoescapes', default='', dest="autoescapes")

    def handle(self, *args, **options):
        translation.activate('en')

        tag_names = options.get('tag_names', '').split(',')
        sources = options.get('sources', '').split(',')
        source_filters = options.get('source_filters', '').split(',')
        params = options.get('params', '').split(',')
        param_filters = options.get('param_filters', '').split(',')
        asvars = options.get('asvars', '').split(',')
        asvar_filters = options.get('asvar_filters', '').split(',')
        blocks = options.get('blocks', '').split(',')
        context_values = options.get('context_values', '').split(',')
        autoescapes = options.get('autoescapes', '').split(',')

        if options['interactive']:
            if tag_names == ['']:
                tag_names = fancy_input("Which tag do you want to use?",
                                        ('t', 't', '{% t ... %}'),
                                        ('ut', 'ut', '{% ut ... %}'))
            if sources == ['']:
                sources = fancy_input(
                    "What type of source string do you want to use?",
                    ('', '<Nothing>', None),
                    ('source', "Variable", '{% t source ... %}'),
                    ('"hello {var}"',
                     "String without XML",
                     '{% t "hello {var}" ... %}'),
                    ('"<xml>hello</xml> {var}"',
                     "String with XML",
                     '{% t "<xml>hello</xml> {var}" ... %}'),
                )
            if source_filters == ['']:
                source_filters = fancy_input(
                    "What filters do you want to apply to the source string?",
                    ('', '<None>', '{% t "hello {var}" ... %}'),
                    ('safe', 'safe', '{% t "hello {var}"|safe ... %}'),
                    ('escape', 'escape', '{% t "hello {var}"|escape ... %}'),
                    ('escapejs', 'escapejs',
                     '{% t "hello {var}"|escapejs ... %}'),
                )
            if params == ['']:
                params = fancy_input(
                    "Do you want to pass a parameter to the tag?",
                    ('var', 'Yes', '{% t "hello {var}" var=var ... %}'),
                    ('', 'No', None),
                )
            if param_filters == ['']:
                param_filters = fancy_input(
                    "What filters do you want to apply to the parameter (if "
                    "you added one)?",
                    ('', '<None>', '{% t "hello {var} var=var ..." %}'),
                    ('safe', 'safe', '{% t "hello {var} var=var|safe ..." %}'),
                    ('escape', 'escape',
                     '{% t "hello {var} var=var|escape ..." %}'),
                    ('escapejs', 'escapejs',
                     '{% t "hello {var} var=var|escapejs ..." %}'),
                )
            if asvars == ['']:
                asvars = fancy_input(
                    "Do you want to save the translation to a context "
                    "variable?",
                    ('', 'No', '{% t ... %}'),
                    ('text', 'Yes', '{% t ... as text %}{{ text }}'),
                )
            if asvar_filters == ['']:
                asvar_filters = fancy_input(
                    "What filters do you want to apply to the saved variable "
                    "(if you chose one)?",
                    ('', '<None>', '{% t ... as text" %}{{ text }}'),
                    ('safe', 'safe', '{% t ... as text" %}{{ text|safe }}'),
                    ('escape', 'escape',
                     '{% t ... as text" %}{{ text|escape }}'),
                    ('escapejs', 'escapejs',
                     '{% t ... as text" %}{{ text|escapejs }}'),
                )
            if blocks == ['']:
                blocks = fancy_input(
                    "What do you want to use as the content of the translated "
                    "block (if no source string was one of the options you "
                    "chose)?",
                    ('', '<Nothing>', None),
                    ('hello {var}',
                     "String without XML",
                     '{% t ... %}hello {var}{% endt %}'),
                    ('<xml>hello</xml> {var}',
                     "String with XML",
                     '{% t ... %}<xml>hello</xml> {var}{% endt %}'),
                )
            if context_values == ['']:
                context_values = fancy_input(
                    "What values do you want to use for your context "
                    "variables?",
                    ('world', 'String without XML', None),
                    ('<xml>world</xml>', 'String with XML', None),
                )
            if autoescapes == ['']:
                autoescapes = fancy_input(
                    "Do you want to render the template using autoescape?",
                    ('yes', 'Yes', None),
                    ('no', 'No', None),
                )

        if autoescapes == ['']:
            autoescapes = ['yes']
        autoescapes = list(set((
            {'yes': True, 'no': False}.get(autoescape.lower(), True)
            for autoescape in autoescapes
        )))

        tests = make_tests(tag_names, sources, source_filters, params,
                           param_filters, asvars, asvar_filters, blocks,
                           context_values)

        print()
        for i, ((template, context), autoescape) in enumerate(
                itertools.product(tests, autoescapes), 1):
            test(template, context, autoescape, i)

        if options['interactive']:
            command = [sys.argv[0], sys.argv[1], sys.argv[2]]
            command.append("--tag-names={}".format(','.join(tag_names)))
            if sources != ['']:
                command.append("--sources='{}'".format(','.join(sources)))
            if source_filters != ['']:
                command.append("--source-filters={}".
                               format(','.join(source_filters)))
            if params != ['']:
                command.append("--params={}".format(','.join(params)))
            if param_filters != ['']:
                command.append("--param-filters={}".
                               format(','.join(param_filters)))
            if asvars != ['']:
                command.append("--asvars={}".format(','.join(asvars)))
            if asvar_filters != ['']:
                command.append("--asvar-filters={}".
                               format(','.join(asvar_filters)))
            if blocks != ['']:
                command.append("--blocks='{}'".format(','.join(blocks)))
            if context_values != ['']:
                command.append("--context-values='{}'".
                               format(','.join(context_values)))
            command.append("--autoescapes={}".format(','.join((
                {True: 'yes', False: 'no'}[autoescape]
                for autoescape in autoescapes
            ))))
            command = ' '.join(command)
            print()
            print("In order to run the same tests again later, use the "
                  "following command:")
            print(command)



from __future__ import absolute_import, unicode_literals

import fnmatch
import io
import os
import sys

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.functional import cached_property
from transifex.common.console import Color
from transifex.native.django.management.common import (NO_LOCALE_DIR,
                                                       TranslatableFile)


class CommandMixin(object):
    """ Common utilities of all subcommands """

    def output(self, msg):
        Color.echo(msg)

    def verbose(self, msg):
        if self.verbose_output:
            Color.echo(msg)

    def _find_files(self, root, subcommand):
        """Get all files in the given root.

        :param basestring root: the root path to search in
        :return: a list of TranslatableFile objects
        :rtype: list
        """
        # TODO: See if we can remove functionality about locale dir and
        # simplify

        def is_ignored(path, ignore_patterns):
            """Check if the given path should be ignored or not."""
            filename = os.path.basename(path)

            def ignore(pattern):
                return (fnmatch.fnmatchcase(filename, pattern) or
                        fnmatch.fnmatchcase(path, pattern))

            return any(ignore(pattern) for pattern in ignore_patterns)

        ignore_patterns = [os.path.normcase(p) for p in self.ignore_patterns]
        dir_suffixes = {'%s*' % path_sep for path_sep in {'/', os.sep}}
        norm_patterns = []
        for p in ignore_patterns:
            for dir_suffix in dir_suffixes:
                if p.endswith(dir_suffix):
                    norm_patterns.append(p[:-len(dir_suffix)])
                    break
            else:
                norm_patterns.append(p)

        all_files = []
        ignored_roots = []
        if self.settings_available:
            ignored_roots = [os.path.normpath(p) for p in
                             (settings.MEDIA_ROOT, settings.STATIC_ROOT) if p]

        follow_links = self.symlinks if subcommand == 'push' else False
        for dirpath, dirnames, filenames in os.walk(root, topdown=True,
                                                    followlinks=follow_links):
            for dirname in dirnames[:]:
                if (
                        is_ignored(os.path.normpath(os.path.join(dirpath,
                                                                 dirname)),
                                   norm_patterns) or
                        os.path.join(os.path.abspath(dirpath), dirname) in
                        ignored_roots):
                    dirnames.remove(dirname)
                    self.verbose('Ignoring directory %s' % dirname)
                elif dirname == 'locale':
                    dirnames.remove(dirname)
                    if subcommand == 'push':
                        self.locale_paths.insert(0, os.path.join(
                            os.path.abspath(dirpath),
                            dirname
                        ))
            for filename in filenames:
                file_path = os.path.normpath(os.path.join(dirpath, filename))
                file_ext = os.path.splitext(filename)[1]
                if (file_ext not in self.extensions or
                        is_ignored(file_path, self.ignore_patterns)):
                    self.verbose('Ignoring file %s in %s' %
                                 (filename, dirpath))
                else:
                    if subcommand == 'push':
                        locale_dir = None
                        for path in self.locale_paths:
                            if os.path.abspath(dirpath).\
                                    startswith(os.path.dirname(path)):
                                locale_dir = path
                                break
                        if not locale_dir:
                            locale_dir = self.default_locale_path
                        if not locale_dir:
                            locale_dir = NO_LOCALE_DIR
                        all_files.append(TranslatableFile(
                            dirpath, filename, locale_dir))
                    elif subcommand == 'migrate':
                        all_files.append(
                            TranslatableFile(dirpath.lstrip('./'), filename)
                        )
        return sorted(all_files)

    def _read_file(self, path, encoding):
        with io.open(path, 'r', encoding=encoding) as fp:
            return fp.read()

    @cached_property
    def settings_available(self):
        try:
            settings.LOCALE_PATHS
        except ImproperlyConfigured:
            if self.verbose_output:
                sys.stderr.verbose("Running without configured settings.\n")
            return False
        return True


def pretty_options(options_dict):
    items = [(k, v) for k, v in options_dict.items()]
    return '\n'.join([' - "{}": {}'.format(x[0], x[1]) for x in items])



import importlib
import re
from datetime import datetime
from hashlib import md5

import pytz


def generate_key(string=None, context=None, parsed=False, plurals=None):
    """Return a unique key based on the given source string and context.

    :param str string: An ICU string
    :param dict plurals: A dictionary with pre-parsed plurals.
        Should be like:
            { 1: "Here is one dog",
              5: "Here are many dogs"}
    :param Union[unicode, list] context: an optional context that
        accompanies the string
    :return: a unique key
    :rtype: str
    """

    def escape_plural(plural):
        """ Escape the : character (and the \ character that will be
        used for escaping)"""
        return plural.replace('\\', '\\\\').replace(':', '\:')

    if not string and not plurals:
        raise ValueError("You need to specify at least "
                         "one of `string`, `plurals`")
    if string and plurals:
        raise ValueError("Cannot use both `string` and `plurals`")

    if not plurals:
        _, plurals = parse_plurals(string)

    string_content = u':'.join(
        u'{}:{}'.format(rule, escape_plural(string))
        for rule, string in sorted(
            plurals.items(), key=lambda x: x[0]
        )
    )

    if context:  # pragma: no cover
        if isinstance(context, list):
            context = u':'.join(context)
        else:
            context = u':'.join(context.split(','))
    if not context:  # pragma: no cover
        context = ''
    return md5(
        (':'.join([string_content, context])).encode('utf-8')
    ).hexdigest()


def now():
    """Return the current datetime, as a UTC-aware object.

    :rtype: datetime
    """
    return datetime.utcnow().replace(tzinfo=pytz.utc)  # pragma no cover


def import_to_python(import_str):
    """Given a string 'a.b.c' return object c from a.b module.

    :param str import_str: a path like a.b.c.
    """
    mod_name, obj_name = import_str.rsplit('.', 1)
    obj = getattr(importlib.import_module(mod_name), obj_name)
    return obj


def make_hashable(data):
    """Make the given object hashable.

    It makes it ready to use in a `hash()` call, making sure that
    it's always the same for lists and dictionaries if they have the same items.

    :param object data: the object to hash
    :return: a hashable object
    :rtype: object
    """
    if isinstance(data, (list, tuple)):
        return tuple((make_hashable(item) for item in sorted(data)))
    elif isinstance(data, dict):
        return tuple(
            (key, make_hashable(value))
            for key, value in sorted(data.items())
        )
    else:
        return data


def parse_plurals(string):
    """ Tries to parse an ICU (possibly pluralized) string, returning its
        plurals separated. It only works if `string` is the simplest possible
        version of a pluralized ICU string, ie whether the plural syntax is on
        the outermost part of the string.

        So, this `{cnt, plural, one {ONE} other {OTHER}}` will return a parsed version,
        but this `hello {cnt, plural, one {ONE} other {OTHER}}` will not.

        The plurals dictionary looks like:
            { 1: "Here is one dog",
              5: "Here are many dogs"}

        If a string cannot be parsed, it's considered a non-pluralized string and is
        assigned the rule "5".

        :rtype: tuple(bool, dict) (Whether the string was parsed & the resulted plurals)
    """

    plurals = {}
    try:
        # {cnt, plural, one {foo} other {foos}}
        # ^^^^^^^^^^^^
        variable_name, remaining = _consume_preamble(string)
        # {cnt, plural, one {foo} other {foos}}
        #               ^^^^^^^^^
        rule, remaining = _consume_rule(remaining)
        plural, remaining = _consume_plural(remaining.strip())
        plurals[rule] = plural
        while remaining.strip():
            # {cnt, plural, one {foo} other {foos}}
            #                         ^^^^^^^^^^^^
            rule, remaining = _consume_rule(remaining.strip())
            plural, remaining = _consume_plural(remaining.strip())
            plurals[rule] = plural
    except Exception:
        return (False, {5: string})

    if ((len(plurals) == 1 and 5 not in plurals) or
            bool({1, 5} - set(plurals.keys()))):
        return (False, {5: string})

    return (True, plurals)


# The `_consume_FOO` functions take an input, "consume" a part of it to produce
# the first return value and return the "unconsumed" part of the input as the
# second return value
def _consume_preamble(string):
    """ Usage:

            >>> _consume_preamble('{cnt, plural, one {ONE} other {OTHER}}')
            <<< ('cnt', 'one {ONE} other {OTHER}')
    """

    if string[0] != '{' or string[-1] != '}':
        raise ValueError()
    first_comma_pos = string.index(',')
    second_comma_pos = string.index(',', first_comma_pos + 1)
    variable_name = string[1:first_comma_pos].strip()
    keyword = string[first_comma_pos + 1:second_comma_pos].strip()
    if not re.search(r'[\w_]+', variable_name) or keyword != "plural":
        raise ValueError()
    return variable_name, string[second_comma_pos + 1:-1].strip()


def _consume_rule(string):
    """ Usage:

            >>> _consume_rule('one {ONE} other {OTHER}')
            <<< ('one', '{ONE} other {OTHER}')
            >>> _consume_rule('other {OTHER}')
            <<< ('other', '{OTHER}')
    """

    def _get_rule_num(rule):
        return {
            'zero': 0, 'one': 1, 'two': 2,
            'few': 3, 'many': 4, 'other': 5
        }[rule]

    left_bracket_pos = string.index('{')
    rule = string[:left_bracket_pos].strip()
    if rule[0] == "=":
        rule = rule[1:]
        rule = int(rule)
        rule = {0: "zero",
                1: "one", 2: "two", 3: "few", 4: "many", 5: "other"}[rule]
    else:
        if rule not in ('zero', 'one', 'two', 'few', 'many', 'other'):
            raise ValueError()

    return _get_rule_num(rule), string[left_bracket_pos:].strip()


def _consume_plural(string):
    """ Usage:

            >>> _consume_plural('{ONE} other {OTHER}')
            <<< ('ONE', 'other {OTHER}')
            >>> _consume_plural('{OTHER}')
            <<< ('OTHER', '')
    """

    bracket_count, escaping = 0, False
    ptr = 0
    while ptr < len(string):
        char = string[ptr]
        if char == "'":
            try:
                peek = string[ptr+1]
            except IndexError:  # pragma: no cover
                peek = None
            if peek == "'":
                ptr += 1
            else:
                if escaping:
                    escaping = False
                else:
                    if peek in ('{', '}'):
                        escaping = True
        elif char == '{':
            if not escaping:
                bracket_count += 1
        elif char == '}':
            if not escaping:
                bracket_count -= 1
            if bracket_count == 0:
                return string[1:ptr], string[ptr + 1:].strip()
        ptr += 1
    raise ValueError()



from transifex.native.core import TxNative


def init(
    token, languages, secret=None,
    cds_host=None, missing_policy=None,
    error_policy=None, cache=None,
):
    """Initialize the framework.

    :param list languages: A list of language codes for the languages
        the application is localized into
    :param str token: the API token to use for connecting to the CDS
    :param str secret: the additional secret required for pushing translations
    :param str cds_host: an optional host for the Content Delivery Service,
        defaults to the host provided by Transifex
    :param AbstractRenderingPolicy missing_policy: an optional policy to use
        for returning strings when a translation is missing
    :param AbstractErrorPolicy error_policy: an optional policy to use
        for defining how to handle translation rendering errors
        :param AbstractCache cache: an optional cache
    """
    if not tx.initialized:
        tx.init(
            languages,
            token,
            secret=secret,
            cds_host=cds_host,
            missing_policy=missing_policy,
            error_policy=error_policy,
            cache=cache,
        )


tx = TxNative()



from __future__ import absolute_import, unicode_literals

from transifex.common.console import Color
from transifex.native import tx
from transifex.native.django.management.utils.base import CommandMixin


class Invalidate(CommandMixin):

    def add_arguments(self, subparsers):
        parser = subparsers.add_parser(
            'invalidate',
            help=('Invalidate CDS, forcing it to re-cache fresh '
                  'translations.'),
        )
        parser.add_argument(
            '--purge', '-p', action='store_true', dest='purge', default=False,
            help=('Force purging CDS cache instead of refreshing content '
                  'asynchronously in the background. (not recommended)'),
        )

    def handle(self, *args, **options):
        purge = options['purge']

        if purge:
            Color.echo('Purging CDS cache...')
        else:
            Color.echo('Invalidating CDS cache...')

        status_code, response_content = tx.invalidate_cache(purge)

        try:
            if 200 <= status_code < 300:
                # {"count":0}
                count = response_content.get('count')
                if purge:
                    Color.echo(
                        '[green]\nSuccessfully purged CDS cache.[end]\n'
                        '[high]Status:[end] [warn]{code}[end]\n'
                        '[high]Records purged: {count}[end]\n'.format(
                            code=status_code,
                            count=count,
                        )
                    )
                else:
                    Color.echo(
                        '[green]\nSuccessfully invalidated CDS cache.[end]\n'
                        '[high]Status:[end] [warn]{code}[end]\n'
                        '[high]Records invalidated: {count}[end]\n'
                        '[high]Note: It might take a few minutes for '
                        'fresh content to be available\n'.format(
                            code=status_code,
                            count=count,
                        )
                    )
            else:
                message = response_content.get('message')
                Color.echo(
                    '[error]\nCould not invalidate CDS.[end]\n'
                    '[high]Status:[end] [warn]{code}[end]\n'
                    '[high]Message:[end] [warn]{message}[end]\n'.format(
                        code=status_code,
                        message=message,
                    )
                )
        except Exception:
            self.output(
                '(Error while printing formatted report, '
                'falling back to raw format)\n'
                'Status: {code}\n'
                'Content: {content}'.format(
                    code=status_code,
                    content=response_content,
                )
            )
